package ru.ifmo.qbeing;

import ru.ifmo.qbeing.common.BehaviourConstants;
import ru.ifmo.qbeing.common.BeingType;
import ru.ifmo.qbeing.global.GlobalStrategy;
import ru.ifmo.qbeing.global.UniverseInfo;
import universum.bi.*;

public class QBeing implements Being, Comparable<QBeing> {

    public static int gcount;
    public static int gdead;

    private QBeing parent;
    private float mass;
    private float speed;
    private float energy;
    private int id;

    private Location location;
    private Integer[] others;

    private QBeingAutomata automata;
    private Location destination;
    private Event action;

    public QBeing() {
        others = new Integer[1];
        automata = new QBeingAutomata(this);
    }

    public float getMass() {
        return mass;
    }

    public void setMass(float mass) {
        this.mass = mass;
    }

    public void reinit(UserGameInfo info) {
        GlobalStrategy.initInstance();
        UniverseInfo.initInstance();
        GlobalStrategy.instance().initGameInfo(info);

        gcount = 0;
        gdead = 0;
    }

    public String getName() {
        return "QBeing3";
    }

    public String getOwnerName() {
        return "Bulemar";
    }

    public BeingParams getParams() {
        return new BeingParams(BehaviourConstants.MAX_MASS, BehaviourConstants.SPEED);
    }

    public synchronized void processEvent(BeingInterface bi, Event e) {
        StateMachineContext ctx;
        switch (e.kind()) {
            case BEING_BORN:
                gcount++;
                ctx = new StateMachineContext();
                ctx.setBeingInterface(bi);
                ctx.setEvent(e);

                EventManager.handleBeingEvent(BeingEvent.E100, this, ctx);
                break;
            case BEING_DEAD:
                gcount--;
                gdead++;

                ctx = new StateMachineContext();
                ctx.setBeingInterface(bi);
                EventManager.handleBeingEvent(BeingEvent.E105, this, ctx);
                bi.log(this, "dead: " + gdead);
                break;
            case BEING_ATTACKED:
                ctx = new StateMachineContext();
                ctx.setBeingInterface(bi);
                ctx.setSender(this);
                float forces = getEnemyForces(bi);
                ctx.put("forces", forces);
                if (getState() == BeingState.S_PRODUCER)
                    EventManager.handleBeingEvent(BeingEvent.E110, this, ctx);
        }
    }

    public Event makeTurn(BeingInterface bi) {
        StateMachineContext machineContext = new StateMachineContext();
        machineContext.setBeingInterface(bi);
        machineContext.setBeing(this);
        EventManager.handleBeingEvent(BeingEvent.E101, this, machineContext);

        return action;
    }

    public QBeing getParent() {
        return parent;
    }

    public void setParent(QBeing parent) {
        this.parent = parent;
    }

    public void setId(int id) {
        this.id = id;
    }

    public int getId() {
        return id;
    }

    public float getEnergy() {
        return energy;
    }

    public void setSpeed(float speed) {
        this.speed = speed;
    }

    public float getSpeed() {
        return speed;
    }

    public Location getLocation() {
        return location;
    }

    public void setLocation(Location location) {
        this.location = location;
    }

    public Integer[] getOthers(BeingInterface bi) {
        others = bi.getPointInfo(this).getEntities(this, others);
        return others;
    }

    public void setEnergy(float energy) {
        this.energy = energy;
    }

    public String toString() {
        return this.getClass().getName();
    }

    public int compareTo(QBeing other) {
        float e1 = getEnergy();
        float e2 = other.getEnergy();
        return -Float.compare(e1, e2);
    }

    public QBeingAutomata getAutomata() {
        return automata;
    }

    public void setDestination(Location destination) {
        this.destination = destination;
    }

    public Location getDestination() {
        return destination;
    }

    public int getState() {
        return automata.getState();
    }

    public float beingWillEat() {
        float willEat = 0;
        if (GlobalStrategy.getType(getState()) == BeingType.TYPE_EXPLORER) {
            if (getEnergy() < BehaviourConstants.NORM_EXP_FOOD * getMass()) {
                willEat = Math.min(getMass() * Constants.K_bite, BehaviourConstants.NORM_EXP_FOOD * getMass() - getEnergy());
            }
        } else if (GlobalStrategy.getType(getState()) == BeingType.TYPE_EATER && getState() == BeingState.S_PRODUCER) {
            willEat = Math.min(getMass() * Constants.K_bite, Constants.K_toborn * getMass() - getEnergy());
        }
        return willEat;
    }

    public void setAction(Event action) {
        this.action = action;
    }

    private Integer[] entitiesHere = new Integer[1];

    private float getEnemyForces(BeingInterface bi) {
        PointInfo pointInfo = bi.getPointInfo(this);
        entitiesHere = pointInfo.getEntities(this, entitiesHere);
        Object owner = GlobalStrategy.instance().getOwner();
        float force = 0;
        for(Integer id: entitiesHere) {
            if (id == null)
                break;
            if (bi.getOwner(this, id) != owner) {
                force += bi.getMass(this, id);
            }
        }
        return force;
    }

    public float getForce() {
        return getEnergy() - getMass() * Constants.K_emin;
    }
}
