package ru.ifmo.qbeing;

import universum.bi.*;

import ru.ifmo.qbeing.global.GlobalStrategy;
import ru.ifmo.qbeing.global.UniverseInfo;
import ru.ifmo.qbeing.common.LocationInfo;
import ru.ifmo.qbeing.common.BehaviourConstants;
import ru.ifmo.qbeing.common.BeingUtil;

import java.util.List;
import java.util.Collection;

public class QBeingAutomata {
    private QBeing being;

    private int state;
    private StateMachineContext context;

    public QBeingAutomata(QBeing being) {
        this.being = being;
        state = BeingState.S_BORN;
    }

// ------------------ Main Automata event processor ------------------------

    public void processBeingEvent(int event, StateMachineContext context) {
        this.context = context;
        switch (state) {
            case BeingState.S_BORN:
                switch(event) {
                    case BeingEvent.E100:
                        z101(); // init being
                        z102(); // ask GlobalStrategy for required being type
                        break;
                    case BeingEvent.E102:
                        state = BeingState.S_PRODUCER;
                        break;
                    case BeingEvent.E103:
                        state = BeingState.S_EXPLORING;
                        break;
                    case BeingEvent.E104:
                        state = BeingState.S_EATER_MOVING_TO_POINT;
                        break;
                }
                break;
            case BeingState.S_EXPLORING:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started
                        break;
                    case BeingEvent.E105:
                        z104();
                        break;
                    case BeingEvent.E106:
                        if (x112()) {
                            z117();
                            state = BeingState.S_EXPLORING_HUNGRY;
                            break;
                        }
                        if (x113()) {
                            z118();
                            state = BeingState.S_PRODUCER;
                            break;
                        }
                        z119();
                        break;
                    case BeingEvent.E107:
                        state = BeingState.S_EATER_FIND_FEED_PLACE;
                        break;
                }
                break;
            case BeingState.S_EXPLORING_HUNGRY:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started
                        break;
                    case BeingEvent.E105:
                        z104(); // inform strategy about being death
                        break;
                    case BeingEvent.E106:
                        if (x114()) {
                            z120();
                            state = BeingState.S_EATER_FIND_FEED_PLACE;
                            break;
                        }
                        if (x110()) {
                            z115();
                            state = BeingState.S_EXPLORING;
                            break;
                        }
                        if (x105()) {
                            z114();
                            break;
                        }
                        if (x111()) {
                            z116();
                            break;
                        }
                        break;
                }
                break;
            case BeingState.S_PRODUCER:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started
                        break;
                    case BeingEvent.E105:
                        z104(); // inform strategy about being death
                        z105(); // remove colony supplier
                        break;
                    case BeingEvent.E103:
                        state = BeingState.S_EXPLORING;
                        break;
                    case BeingEvent.E106:
                        if (x102()) {
                            z108();
                            break;
                        }
                        if (x103()) {
                            z109(); // Transform colony to explorers
                            break;
                        }
                        if (x104()) {
                            z111(); // Attack
                            break;
                        }
                        z110();
                        break;
                    case BeingEvent.E110:
                        if (x115()) {
                            z121();
                            break;
                        }
                        break;
                    case BeingEvent.E111:
                        state = BeingState.S_SUPPORT;
                        break;
                }
                break;
            case BeingState.S_EATER_MOVING_TO_POINT:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started

                        break;
                    case BeingEvent.E105:
                        z104(); // inform strategy about being death
                        z103(); // Inform strategy that being haven't reached it's destination
                        break;
                    case BeingEvent.E106:
                        if (x101()) {
                            state = BeingState.S_PRODUCER;
                            z106();
                            break;
                        }
                        z107();
                        break;
                }
                break;
            case BeingState.S_SUPPORT:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started
                        break;
                    case BeingEvent.E105:
                        z104(); // inform strategy about being death
                        z103(); // Inform strategy that being haven't reached it's destination
                        break;
                    case BeingEvent.E106:
                        if (x101()) {
                            state = BeingState.S_PRODUCER;
                            z122();
                            break;
                        }
                        z107();
                        break;
                }
                break;
            case BeingState.S_EATER_FIND_FEED_PLACE:
                switch(event) {
                    case BeingEvent.E101:
                        z100(); // new move started
                        break;
                    case BeingEvent.E105:
                        z104(); // inform strategy about being death
                        break;
                    case BeingEvent.E106:
                        if (x105()) {
                            z112();
                            state = BeingState.S_PRODUCER;
                            break;
                        }
                        if (x109() || x106()) {
                            z113();
                            state = BeingState.S_EATER_MOVING_TO_POINT;
                            break;
                        }
                        if (x107()) {
                            z114();
                            break;
                        }
                        if (x108()) {
                            z113();
                            state = BeingState.S_EATER_MOVING_TO_POINT;
                            break;
                        }

                        break;
                }
                break;
        }
    }

// ------------------ End of Main Automata event processor ---------

// --------------- Automata input actions --------------------------

    /**
     * We have reached the destination point
     * @return true, if we reched destination
     */
    private boolean x101() {
        BeingInterface bi = context.getBeingInterface();
        QBeing me = context.getBeing();
        return bi.getLocation(me).equals(me.getDestination());
    }

    /**
     * Being can born
     * @return true, if being can born
     */
    private boolean x102() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        return bi.getEnergy(me) >= Constants.K_toborn * me.getMass();
    }

    /**
     * Lack of food, should tranform the colony to explorers
     * @return true, if food amount is not enough
     */
    private boolean x103() {
        return lackOfFood() && !GlobalStrategy.instance().isExploringFinished();
    }

    /**
     * Is there an enemy we should attack?
     * @return true, if the enemy is located, and we are healthy enough to attack.
     */
    private boolean x104() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Integer other = isForeignHere(me, bi);
        float e = bi.getEnergy(me);
        context.put("enemy", other);
        return other != null && e >= BehaviourConstants.MIN_EATER_ATTACK_ENERGY * me.getMass();
    }

    /**
     * We have enough energy at the current location.
     * @return true, if there is enough food.
     */
    private boolean x105() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        return UniverseInfo.instance().isEnoughEnergy(me, new LocationInfo(bi.getPointInfo(me), me), BehaviourConstants.MIN_FEED_ENERGY);
    }

    /**
     * Is there an energy point with enogh energy to start a colony we can reach
     * @return true, if there is one
     */
    private boolean x106() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.MIN_START_COLONY_ENERGY);
        context.put("location", loc);
        return loc != null;
    }

    /**
     * We are dying
     * @return true, if we have low energy
     */
    private boolean x107() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        return bi.getEnergy(me) < BehaviourConstants.MIN_HEALTH * me.getMass();
    }

    /**
     * Is there an energy point with some energy we can reach
     * @return true, if there is one
     */
    private boolean x108() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.MIN_FEED_ENERGY);
        context.put("location", loc);
        return loc != null;
    }

    /**
     * We can reach the energy point with energy
     * @return true, if we can
     */
    private boolean x109() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.MIN_FEED_ENERGY);
        if (loc != null) {
            float avail = UniverseInfo.instance().getAvailEnergyAtLoc(me, loc);
            me.setDestination(loc);
            if (avail > BehaviourConstants.CHANGE_FEED_PLACE_REQUIRED) {
                context.put("location", loc);
                return true;
            }
        }
        return false;
    }

    /**
     * We re not hugry anymore
     * @return true, if we are healthy enough to explore.
     */
    private boolean x110() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        return bi.getEnergy(me) >= BehaviourConstants.NORM_EXP_FOOD * me.getMass();
    }

    /**
     * Is there an energy point with some energy we can reach (Greedy)
     * @return true, if there is one
     */
    private boolean x111() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.MIN_FEED_ENERGY);
        if (loc == null) {
            loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.MIN_FEED_ENERGY, true);
        }
        context.put("location", loc);
        return loc != null;
    }

    /**
     * Explorer is hungry.
     * @return
     */
    private boolean x112() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        return bi.getEnergy(me) < BehaviourConstants.MIN_EXP_FOOD * me.getMass();
    }

    /**
     * Can start colony.
     * @return
     */
    private boolean x113() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        PointInfo pi = bi.getPointInfo(me);
        return UniverseInfo.instance().isEnoughEnergy(me, new LocationInfo(pi, me),
                BehaviourConstants.MIN_START_COLONY_ENERGY) && GlobalStrategy.instance().startColonyHere(pi, me);
    }

    /**
     * Explorer should be set to Eater.
     * @return true, if there is a point with enough food
     */
    private boolean x114() {
        return x110() && GlobalStrategy.instance().isExploringFinished();
    }

    /**
     * We need support
     * @return true, if support is needed
     */
    private boolean x115() {
        Float forces = (Float) context.get("forces");
        context.setBeing(being);
        QBeing me = context.getBeing();
        float supportNeeded = forces * BehaviourConstants.ENERGY_FROM_MASS - GlobalStrategy.instance().getColonyForce(me);
        if (supportNeeded > 0) {
            context.put("supportNeeded", supportNeeded);
            return true;
        }
        return false;
    }
//  ------------ End of Automata input actions -------------

//  ------------ Automata output actions -------------------

    /**
     * Update being energy, location, etc.
     */
    private void z100() {
        being.setAction(null);
        BeingInterface bi = context.getBeingInterface();
        being.setLocation(bi.getLocation(being));
        GlobalStrategy.instance().updateEnergy(being, bi.getEnergy(being));
        UniverseInfo.instance().updateKnownLocations(being, bi);
        StateMachineContext ctx = new StateMachineContext();
        ctx.setBeingInterface(bi);
        EventManager.handleStrategyEvent(StrategyEvent.E200, ctx);
        EventManager.handleBeingEvent(BeingEvent.E106, being, context);
    }

    /**
     * Init being
     */
    private void z101() {
        BeingInterface bi = context.getBeingInterface();
        Event e = context.getEvent();
        if (e.param() != null) {
            BeingParams bp = (BeingParams) e.param();
            being.setMass(bp.M);
            being.setSpeed(bp.S);
            being.setEnergy(bi.getEnergy(being));
            being.setParent((QBeing) bp.parameter);
            being.setId(bi.getId(being));
            GlobalStrategy.instance().setOwner(bi.getOwner(being, bi.getId(being)));
        }
    }

    /**
     * Ask GlobalStrategy for required being type
     */
    private void z102() {
        BeingInterface bi = context.getBeingInterface();
        StateMachineContext ctx = new StateMachineContext();
        ctx.setSender(being);
        ctx.setBeingInterface(bi);
        EventManager.handleStrategyEvent(StrategyEvent.E201, ctx);
    }

    /**
     * Inform strategy that being haven't reached it's destination
     */
    private void z103() {
        UniverseInfo.instance().removeBeingFollowingLocation(being, being.getDestination());
        UniverseInfo.instance().removeSupportBeing(being, being.getDestination());
    }

    /**
     * Inform strategy about being death
     */
    private void z104() {
        StateMachineContext context = new StateMachineContext();
        context.setBeing(being);
        EventManager.handleStrategyEvent(StrategyEvent.E202, context);
    }

    /**
     * Inform strategy of being death.
     */
    private void z105() {
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().colonySuplierRemoved(bi.getLocation(being), being);
    }

    /**
     * Eater reached its destination point
     */
    private void z106() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        UniverseInfo.instance().removeBeingFollowingLocation(me, me.getDestination());
        GlobalStrategy.instance().changeState(BeingState.S_EATER_MOVING_TO_POINT, BeingState.S_PRODUCER, me, bi);
    }

    /**
     * Step towards destination point
     */
    private void z107() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        being.setAction(GlobalStrategy.instance().moveTo(me, bi, me.getDestination()));
    }

    /**
     * Born
     */
    private void z108() {
        QBeing me = context.getBeing();
        BeingParams bp = me.getParams();
        bp.parameter = me;
        being.setAction(new Event(EventKind.ACTION_BORN, bp));
    }

    /**
     * Transform colony to explorers
     */
    private void z109() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().transformColonyToExplorers(me, bi);
    }

    /**
     * Process colony advice
     */
    private void z110() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        being.setAction(GlobalStrategy.instance().colonyAction(bi, me));
    }

    /**
     * Attack enemy.
     */
    private void z111() {
        being.setAction(new Event(EventKind.ACTION_ATTACK, context.get("enemy")));
    }

    /**
     * Start eating.
     */
    private void z112() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().changeState(BeingState.S_EATER_FIND_FEED_PLACE, BeingState.S_PRODUCER, me, bi);
    }

    /**
     * Go to energy point
     */
    private void z113() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location loc = (Location) context.get("location");
        UniverseInfo.instance().setBeingFollowingLocation(me, loc);
        me.setDestination(loc);
        GlobalStrategy.instance().changeState(BeingState.S_EATER_FIND_FEED_PLACE, BeingState.S_EATER_MOVING_TO_POINT, me, bi);
    }

    /**
     * Eat.
     */
    private void z114() {
        QBeing me = context.getBeing();
        being.setAction(new Event(EventKind.ACTION_EAT, me.getMass()));
    }

    /**
     * Continue exploring.
     */
    private void z115() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().changeState(BeingState.S_EXPLORING_HUNGRY, BeingState.S_EXPLORING, me, bi);
    }

    /**
     * Explorer Step towards energy point
     */
    private void z116() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        being.setAction(GlobalStrategy.instance().moveTo(me, bi, (Location) context.get("location")));
    }

    /**
     * Set Explorer to hungry state
     */
    private void z117() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().changeState(BeingState.S_EXPLORING, BeingState.S_EXPLORING_HUNGRY, me, bi);
    }

    /**
     * Start colony
     */
    private void z118() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().changeState(BeingState.S_EXPLORING, BeingState.S_PRODUCER, me, bi);
    }

    /**
     * Continue exploring
     */
    private void z119() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Location nextLocation;
        // move to one of the neighbour point to set colony
        Location enoughEnergyLocation = UniverseInfo.instance().findEnoughEnergy(me, bi.getNeighbourInfo(me), BehaviourConstants.MIN_START_COLONY_ENERGY);
        if (enoughEnergyLocation != null) {
            nextLocation = enoughEnergyLocation;
        } else {
            // continue search
            nextLocation = calculateNextRadiusLocation(bi, me);
        }
        being.setAction(GlobalStrategy.instance().moveTo(me, bi, nextLocation));
    }

    /**
     * Set Eater being to "Find food" state
     */
    private void z120() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        GlobalStrategy.instance().changeState(BeingState.S_EXPLORING_HUNGRY, BeingState.S_EATER_FIND_FEED_PLACE, me, bi);
    }

    /**
     * Call support
     */
    private void z121() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        Float supportNeeded = (Float) context.get("supportNeeded");

        GlobalStrategy.instance().getSupport(bi, me.getLocation(), supportNeeded);
    }

    /**
     * Support reached its destination point
     */
    private void z122() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        UniverseInfo.instance().removeSupportBeing(me, me.getDestination());
        GlobalStrategy.instance().changeState(BeingState.S_SUPPORT, BeingState.S_PRODUCER, me, bi);
    }
//  ------------ End of Automata output actions -------------

//  ------------- Misc. Methods ------------
    /**
     * Returns current state
     * @return state
     */
    public int getState() {
        return state;
    }

    private boolean lackOfFood() {
        QBeing me = context.getBeing();
        BeingInterface bi = context.getBeingInterface();
        PointInfo pi = bi.getPointInfo(me);
        float avail = UniverseInfo.instance().getAvailEnergy(me, new LocationInfo(pi, me));
        return avail < BehaviourConstants.MIN_FEED_ENERGY ;
    }

    private Integer isForeignHere(QBeing me, BeingInterface bi) {
        Object owner = GlobalStrategy.instance().getOwner();
        Integer[] others = me.getOthers(bi);
        for (Integer other : others) {
            if (other == null) {
                break;
            }
            if (owner != bi.getOwner(me, other)) {
                return other;
            }
        }
        return null;
    }

    private Location calculateNextRadiusLocation(BeingInterface bi, QBeing me) {
        Location resultLocation;
        List<Location> reachable = bi.getReachableLocations(me);
        Location center = UniverseInfo.instance().getExploreCenters(me);
        Collection<Location> notVisited = UniverseInfo.instance().filterVisited(bi, reachable);
        if (notVisited.isEmpty()) {
            resultLocation = BeingUtil.getRandomFarest(bi, center, reachable);
        } else {
            resultLocation = BeingUtil.getRandomNearest(bi, center, notVisited);
        }
        return resultLocation;
    }
}
